'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command import git_clone
from watchme.logger import bot

def main(args, extra):
    '''get a watcher using git, meaning clone to a temporary location and then
       copying the entire repo (or a subfolder) to the watcher base.
    '''    

    # Required - will print help if not provided
    repo = args.repo[0]
    
    # If a custom name is provided, add it
    if extra != None:
        extra = extra.pop(0)

    # Clone the watcher, and optionally just one task
    git_clone(repo=repo, base=args.base, name=extra, force=args.force)




'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.utils import (
    write_file,
    write_json
)

import shutil
import os

class TaskBase(object):

    def __init__(self, name, params={}, **kwargs):

        # Ensure subclass was created correctly
        for req in ['required_params', 'type', 'export_func']:
            if not hasattr(self, req):
                bot.exit('A Task must have a %s function or attribute.' % req)

        self.name = name
        self.valid = False
        self.params = {}
        self.set_params(params)
        self.validate()

    def get_type(self):
        '''get the watcher type.
        '''
        return self.type


# Identification

    def __repr__(self):
        return "[task|%s]" %self.name

    def __str__(self):
        return "[task|%s]" %self.name


# Parameters

    def set_params(self, params):
        '''iterate through parameters, set into dictionary.

           Parameters
           ==========
           params: a list of key@value pairs to set.
        '''
        for key,value in params.items():
            key = key.lower()
            self.params[key] = value


    def export_params(self, active="true"):
        '''export parameters, meaning returning a dictionary of the task
           parameters plus the addition of the task type and active status.
        '''
        params = self.params.copy()
        params['active'] = active
        params['type'] = self.type
        return params


# Validation

    def validate(self):
        '''validate the parameters set for the Task. Exit if there are any
           errors. Ensure required parameters are defined, and have correct
           values.
        '''
        self.valid = True

        for param in self.required_params:
            if param not in self.params:
                bot.error('Missing required parameter: %s' % param)
                self.valid = False

        # Call subclass validation function
        self._validate()

    def _validate(self):
        '''validation function intended to be implemented by subclass.
        '''
        pass


# Run Single Task

    def run(self):
        '''run an isolated task, meaning no update or communication with
           the watcher. This will return the raw result.
        '''
        params = self.export_params()
        func = self.export_func()
        if func != None:
            return func(**params)
        bot.error('Cannot find function.')



# Saving


    def _save_list(self, results, repo, func, file_name):
        '''general function to perform saving a list of content to json,
           text, or moving paths. Used by save_file_list, save_json_list,
           and save_text_list.

           Parameters
           ==========
           results: list of results, assumed to be the correct type
           repo: the repository base with the task folder
           func: the function to call.
           file_name: the base file name to use.
        '''
        file_name, ext = os.path.splitext(file_name)
        files = []

        for r in range(len(results)):
            result = results[r]
            filename = "%s-%s%s" %(file_name, str(r), ext)
            saved = func(result, repo, filename)
            files.append(saved)

        return files

    def _save_text(self, result, repo, file_name=None):
        '''save a general text object to file.
 
           Parameters
           ==========
           result: the result object to save, should be path to a file
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', file_name) or 'result.txt'
        task_folder = os.path.join(repo, self.name)
        destination = os.path.join(task_folder, file_name)
        write_file(destination, result)
        return destination

    def _save_file(self, result, repo, file_name=None):
        '''for a result that exists, move the file to final destination.
 
           Parameters
           ==========
           result: the result object to save, should be path to a file
           repo: the repository base with the task folder
        '''
        if os.path.exists(result):
            name = os.path.basename(result)
            file_name = self.params.get('file_name', file_name) or name
            task_folder = os.path.join(repo, self.name)

            # HOME/.watchme/watcher/<task>/<result>
            destination = os.path.join(task_folder, file_name)
            shutil.move(result, destination)

            # <task>/<result>     
            return os.path.join(name, file_name)

        bot.warning('%s does not exist.' % result)
        
    def _save_json(self, result, repo, file_name=None):
        '''for a result that is a dictionary or list, save as json
 
           Parameters
           ==========
           result: the result object to save, should be dict or list
        '''
        file_name = self.params.get('file_name', file_name) or 'result.json'
        destination = os.path.join(repo, self.name, file_name)
        write_json(result, destination)
        return destination


    def _save_json_list(self, results, repo):
        '''A wrapper around save json for a list, handles file naming.
 
           Parameters
           ==========
           results: the results to save, should a list
        '''
        file_name = self.params.get('file_name', 'result.json')
        return self._save_list(results, repo, self._save_json, file_name)


    def _save_file_list(self, results, repo):
        '''for a list of results that exist, move the files to 
           final destination.
 
           Parameters
           ==========
           results: list of paths to a file, those not existing are skipped
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', os.path.basename(result))
        return self._save_list(results, repo, self._save_file, file_name)


    def _save_text_list(self, results, repo):
        '''for a list of general text results, write them to output files.
 
           Parameters
           ==========
           results: list of string results to write to file
           repo: the repository base with the task folder
        '''
        file_name = self.params.get('file_name', 'result.txt')
        return self._save_list(results, repo, self._save_text, file_name)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''run a watcher, optionally supplying one or more regular expressions to
       check for.
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    watcher = get_watcher(name, base=args.base, create=False)
    
    # If a set of task names or regular expressions are provided:
    if extra != None:
        extra = "(%s)" % '|'.join(extra)

    # Run the watcher, providing regular expressions to match tasks
    watcher.run(regexp=extra, 
                parallel=not args.serial, 
                test=args.test,
                show_progress=not args.no_progress)




'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.logger import bot

def main(args, extra):
    '''activate one or more watchers
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    task = args.task[0]

    if not task.startswith('task'):
        example = 'watchme add watcher task-cpu func@cpu_task type@psutils'
        bot.exit('Task name must start with "task", e.g., %s' % example)

    # Exit if the user doesn't provide any parameters
    if extra == None:
        bot.exit('Please provide parameters to add to your watcher (key@value)')
              
    # Type can also be an argument
    watcher_type = args.watcher_type
    params = []
    for param in extra:
       if param.startswith('type@'):
           watcher_type = param.replace('type@', '')
       else:
           params.append(param)

    # Get the watcher to interact with, must already exist
    watcher = get_watcher(name, base=args.base, create=False) 

    # Add the task. Will exit if not a valid type, or parameters
    watcher.add_task(task=task,
                     task_type=watcher_type,
                     params=params,
                     force=args.force,
                     active=args.active)




# Copyright (C) 2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


__version__ = "0.0.13"
AUTHOR = 'Vanessa Sochat'
AUTHOR_EMAIL = 'vsochat@stanford.edu'
NAME = 'watchme'
PACKAGE_URL = "http://www.github.com/vsoch/watchme"
KEYWORDS = 'web, changes, cron'
DESCRIPTION = "client to watch for webpage changes, and track over time"
LICENSE = "LICENSE"

INSTALL_REQUIRES = (
    ('python-crontab', {"min_version": "2.3.6"}),
    ('configparser', {"min_version": "3.5.3"}),
)

## beautiful soup selection task

INSTALL_URLS_DYNAMIC = (
    ('beautifulsoup4', {'min_version': '4.6.0'}),
)

INSTALL_PSUTILS = (
    ('psutil', {'min_version': '5.4.3'}),
)

INSTALL_ALL = (INSTALL_REQUIRES +
               INSTALL_PSUTILS +
               INSTALL_URLS_DYNAMIC)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.defaults import ( 
    WATCHME_WATCHER, 
    WATCHME_BASE_DIR,
    WATCHME_DEFAULT_TYPE
)
from watchme.utils import ( 
    get_installdir, 
    read_file 
)
import configparser
import getpass
import shutil
import os


# CONFIG TEMPLATES #############################################################

def get_configfile_template():
    '''return the full path to the default configuration file
    '''
    return _get_config('watchme.cfg')

def _get_config(name):
    '''shared function to return a file in the config directory
    '''
    return os.path.abspath(os.path.join(get_installdir(), 'config', name))


# ACTIVE CONFIG FILES ##########################################################

def get_configfile(name, base=None):
    '''return the full path to a specific watcher configuration
    '''
    if base == None:
        base = WATCHME_BASE_DIR

    configfile = os.path.join(base, name, 'watchme.cfg')
    check_exists(filename)
    return configfile
    

# CONFIG IO ####################################################################

def write_config(filename, config, mode="w"):
    '''use configparser to write a config object to filename
    '''
    with open(filename, mode) as filey:
        config.write(filey)
    return filename


def read_config(filename):
    '''use configparser to write a config object to filename
    '''
    check_exists(filename)
    config = configparser.ConfigParser()
    config.read(filename)
    return config


# WATCHER CONFIG ###############################################################

def generate_watcher_config(path, watcher_type=None):
    '''generate a watcher config, meaning a watcher folder in the watchme
       base folder.

       Parameters
       ==========
       path: the path to the watcher repository
    '''
    check_exists(path)
    configfile = get_configfile_template()
    watcher_config = os.path.join(path, 'watchme.cfg')
    if not os.path.exists(watcher_config):
        bot.info('Generating watcher config %s' % watcher_config)
        shutil.copyfile(configfile, watcher_config)

    # Complete generation includes the watcher type
    if watcher_type == None:
        watcher_type = WATCHME_DEFAULT_TYPE
    
    # The template config has the section, but just in case
    config = read_config(configfile)
    if 'watcher' not in config.sections():
        config.add_section('watcher')
    config['watcher']['type'] = watcher_type

    # Save to file
    write_config(configfile, config)

# CONFIG HELPERS ###############################################################

def check_exists(filename):
    '''a general helper function to check for existence, and exit if not found.
    '''
    if not os.path.exists(filename):
        bot.exit('Cannot find %s' % filename)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''


from watchme.logger import bot
from subprocess import (
    Popen,
    PIPE,
    STDOUT
)
import os
import re
import shlex


# User Prompts

def confirm_prompt(prompt):
    '''A wrapper around choice_prompt, but ask the user specifically for a
       yes / no response that is converted to boolean for the calling agent.

       Parameters
       ==========
       prompt: the prompt to ask the user    
    '''
    choice = choice_prompt(prompt, choices = ["Y", "N", "y", "n"])
    return convert2boolean(choice)


def choice_prompt(prompt, choices=None, choice=None):
    '''Ask the user for a prompt, and only return when one of the requested
       options is provided.

       Parameters
       ==========
       prompt: the prompt to ask the user
       choices: a list of choices that are valid, defaults to [Y/N/y/n]
    
    '''
    if not choices:
        choices = ["y", "n", "Y", "N"]

    print(prompt)
    get_input = getattr(__builtins__, 'raw_input', input)
    pretty_choices = '/'.join(choices)
    message = 'Please enter your choice [%s] : ' %(pretty_choices)
    while choice not in choices:
        choice = get_input(message).strip()

        # If the option isn't valid, this is shown next
        message = "Please enter a valid option in [%s]" %(pretty_choices)    
    return choice


def regexp_prompt(prompt, regexp='.', answer=''):
    '''Ask the user for a text entry that matches a regular expression

       Parameters
       ==========
       prompt: the prompt to ask the user
       regexp: the regular expression to match. defaults to anything.
    
    '''
    get_input = getattr(__builtins__, 'raw_input', input)
    while not re.search(regexp, answer):
        answer = get_input(prompt + ': ').strip()
        # If the option isn't valid, this is shown next
        message = "Your entry must match the regular expression %s" % regexp    

    return answer


# Terminal Commands

def which(software, strip_newline=True):
    '''get_install will return the path to where an executable is installed.
    '''
    if software is None:
        software = "watchme"
    cmd = 'which %s' % software
    try:
        result = run_command(cmd)
        if strip_newline is True:
            result['message'] = result['message'].strip('\n')
        if "message" in result:
            return result['message']
        return result

    except: # FileNotFoundError
        return None


def get_installdir():
    '''get_installdir returns the installation directory of the application
    '''
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def run_command(cmd, sudo=False):
    '''run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    '''

    cmd = shlex.split(cmd)

    if sudo is True:
        cmd = ['sudo'] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd,stderr=STDOUT,stdout=PIPE)

    t = output.communicate()[0],output.returncode
    output = {'message':t[0],
              'return_code':t[1]}

    if isinstance(output['message'], bytes):
        output['message'] = output['message'].decode('utf-8')

    return output


def convert2boolean(arg):
    '''
    convert2boolean is used for environmental variables
    that must be returned as boolean
    '''
    if not isinstance(arg, bool):
        return arg.lower() in ("yes", "true", "t", "1", "y")
    return arg



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.command import (
    get_watchers,
    list_watcher
)
from watchme.logger import bot

def main(args, extra):
    '''list installed watchers
    '''    
    if extra == None:
        get_watchers(args.base)
    else:
        for watcher in extra:
            list_watcher(watcher, args.base)




#!/usr/bin/python

# Copyright (C) 2019 Vanessa Sochat.

# This Source Code Form is subject to the terms of the
# Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest.mock import patch
import unittest
import tempfile
import shutil
import json
import os


print("############################################################ test_utils")

class TestUtils(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        
    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        

    def test_write_read_files(self):
        '''test_write_read_files will test the functions write_file and read_file
        '''
        print("Testing utils.write_file")
        from watchme.utils import write_file
        import json
        tmpfile = tempfile.mkstemp()[1]
        os.remove(tmpfile)
        write_file(tmpfile,"mocos!")
        self.assertTrue(os.path.exists(tmpfile))        

        print("Testing utils.read_file...")
        from watchme.utils import read_file
        content = read_file(tmpfile)[0]
        self.assertEqual("mocos!", content)

        from watchme.utils import write_json
        print("Testing utils.write_json.")
        print("...Case 1: Providing bad json")
        bad_json = {"IWuvWaffles?'}":[{True}, "2", 3]}
        tmpfile = tempfile.mkstemp()[1]
        os.remove(tmpfile)        
        with self.assertRaises(TypeError) as cm:
            write_json(bad_json, tmpfile)

        print("...Case 2: Providing good json")        
        good_json = {"IWuvWaffles!": [True, "2", 3]}
        tmpfile = tempfile.mkstemp()[1]
        os.remove(tmpfile)
        write_json(good_json,tmpfile)
        with open(tmpfile,'r') as filey:
            content = json.loads(filey.read())
        self.assertTrue(isinstance(content, dict))
        self.assertTrue("IWuvWaffles!" in content)

        print("Testing utils.print_json")
        from watchme.utils import print_json
        result=print_json({1:1})
        self.assertEqual('{\n    "1": 1\n}', result)


    def test_get_installdir(self):
        '''get install directory should return the base of where singularity
        is installed
        '''
        print("Testing utils.get_installdir")
        from watchme.utils import get_installdir
        whereami = get_installdir()
        print(whereami)
        self.assertTrue(whereami.endswith('watchme'))

    def test_prompts(self):
        '''test user inputs and prompts'''
        from watchme.utils import confirm_prompt

        print('Testing utils.confirm_prompt')
        user_input = [
            'Y',
            'y',
            'n',
            'N',
        ]
        expected_responses = [
             True,
             True,
             False,
             False
        ]
        for i in range(len(user_input)):
            with patch('builtins.input', side_effect=user_input[i]):
                response = confirm_prompt("Please confirm this thing.")
            self.assertEqual(response, expected_responses[i])

    def test_terminal(self):

        print('Testing utils.run_command')
        from watchme.utils import run_command
        from watchme.utils import which
        result = run_command('echo Las Papas Fritas')
        self.assertEqual(result['message'], 'Las Papas Fritas\n')
        self.assertEqual(result['return_code'], 0)     

        print('Testing utils.which')
        result = which('echo')
        self.assertEqual(result, '/bin/echo')

    def test_userhome(self):
        print('Testing utils.get_user')
        print('Testing utils.get_userhome')
        from watchme.utils import get_userhome
        from watchme.utils import get_user
        user = get_user()
        userhome = get_userhome()
        self.assertEqual('/home/%s' % user, userhome)

    def test_files(self):
        print('Testing utils.generate_temporary_files')
        from watchme.utils import generate_temporary_file

        tmpfile = generate_temporary_file()
        from watchme.utils import run_command
        run_command('touch %s' %tmpfile)
        print('Testing utils.copyfile')
        from watchme.utils import copyfile
        copyfile(tmpfile, '%s.booga' % tmpfile)
        self.assertTrue(os.path.exists('%s.booga' %tmpfile))
        self.assertTrue(os.path.exists(tmpfile))
        os.remove(tmpfile)
        os.remove("%s.booga" % tmpfile) 

    def test_mkdir(self):
        print('Testing utils.mkdir_p')
        from watchme.utils import mkdir_p
        tmpdir = os.path.join(self.tmpdir, "if", "I", "were", "an", "avocado")
        mkdir_p(tmpdir)
        self.assertTrue(os.path.exists(tmpdir))

        print('Testing utils.get_tmpdir')
        from watchme.utils import get_tmpdir
        tmpdir = get_tmpdir()
        self.assertTrue(os.path.exists(tmpdir))
        self.assertTrue('watchme' in tmpdir)
        shutil.rmtree(tmpdir)


if __name__ == '__main__':
    unittest.main()



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from crontab import CronTab
from watchme.utils import ( which, get_user )

# Scheduling


def remove_schedule(self, name=None, quiet=False):
    '''remove a scheduled item from crontab, this is based on the watcher
       name. By default, we use the watcher instance name, however you
       can specify a custom name if desired.
    '''
    if name == None:
        name = self.name

    cron = self.get_crontab()

    comment = 'watchme-%s' % self.name
    found = False
    for job in cron.find_comment(comment):
        found = True
        cron.remove(job)
    
    if found is True:
        bot.info('Removed schedule for watcher %s' % name)
        cron.write_to_user(user=True)
    return cron


def has_schedule(self, must_exist=False):
    '''determine if a watcher already has a schedule, as a warning to the user.
    '''
    if self.get_job(must_exist=False) == None:
        return False
    return True


def get_job(self, must_exist=True):
    '''return the job to the user, or None
    '''
    # Find the job based on a standard of comment
    cron = self.get_crontab()
    comment = 'watchme-%s' % self.name
    job = list(cron.find_comment(comment)) or [None]
    job = job[0]

    # Return None, or the actual cron job
    if job == None and must_exist == True:
        bot.warning('%s does not have a cron job configured' % self.name)
    
    return job


def get_crontab(self):  
    '''get an instance of the user's crontab. We use the running user.
    '''
    # Create an instance of the user's crontab
    return CronTab(user=True)


def update_schedule(self, minute=12, hour='*', month='*', day='*'):
    '''update a scheduled item from the crontab, with a new entry. This
       first looks for the entry (and removes it) and then clls the new_
       schedule function to write a new one. This function is intended
       to be used by a client from within Python, and isn't exposed from
       the command line.
    '''
    job = self.get_job(must_exist=True)

    # Update the job if found
    if job != None:
        return self.schedule(minute=minute,
                             hour=hour,
                             month=month, 
                             day=day,
                             job=job,
                             force=True)

    bot.warning('%s does not have a current schedule to update.' % self.name)

def clear_schedule(self):
    '''clear all cron jobs associated with the watcher. To remove jobs
       associated with a single watcher, use remove_schedule
    '''
    cron = self.get_crontab()
    bot.info('Clearing jobs associated with all watchers')
    cron.remove_all(comment='watchme-*')

    # Save new cron
    cron.write_to_user(user=True)
    return cron


def schedule(self, 
             minute=12, 
             hour=0, 
             month='*', 
             day='*', 
             weekday='*',
             job=None,
             force=False):
    '''schedule the watcher to run at some frequency to update record of pages.
       By default, the task will run at 12 minutes passed midnight, daily.
       You can change the variables to change the frequency. See
       https://crontab.guru/ to get a setting that works for you.

            Hourly:	0 * * * *
            Daily:	0 0 * * *    (midnight) default
            weekly	0 0 * * 0
            monthly	0 0 1 * *
            yearly	0 0 1 1 *

       Parameters
       ==========
       minute: must be within 1 and 60, or set to "*" for every minute
       hour: must be within 0 through 23 or set to *
       month: must be within 1 and 12, or *
       day: must be between 1 and 31, or *
       weekday: must be between 0 and 6 or *
       job: if provided, assumes we are updated an existing entry.
    '''
    cron = self.get_crontab()

    # Cut out early if the job already exists, and force is false
    if self.has_schedule() and not force:
        bot.exit('%s already has a schedule. Use --force to update.' % self.name)

    # Remove any previous schedules
    cron = self.remove_schedule(quiet=True)

    # minute must be between * or 0 through 59, or *
    if minute not in ['*'] + list(range(60)):
        bot.exit('minute must be in [0..59] or equal to *')

    # Hour must be between 0 through 23, or *
    if hour not in ['*'] + list(range(24)):
        bot.exit('hour must be in [0..23] or equal to *')
  
    # Day must be in range 1 through 31, or *
    if day not in ['*'] + list(range(1,32)):
        bot.exit('day must be in [1..31] or equal to *')

    # Day must be in range 1 through 31, or *
    if month not in ['*'] + list(range(1,13)):
        bot.exit('month must be in [1..12] or equal to *')

    # Day must be in range 1 through 31, or *
    if weekday not in ['*'] + list(range(7)):
        bot.exit('weekday must be in [0..6] or equal to *')

    # The command will run the watcher, watcher.cfg controls what happens
    whereis = which('watchme')
    command = '%s run %s' % (whereis, self.name)
    comment = 'watchme-%s' % self.name

    if job == None:
        job  = cron.new(command=command, comment=comment)

    # Set the time, and then write the job to file
    job.setall(minute, hour, day, month, weekday)
    job.enable()
    cron.write_to_user(user=True)
    bot.info(job)    

    return job



#!/usr/bin/env python

'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import bot
from watchme.defaults import ( 
    WATCHME_WATCHER, 
    WATCHME_TASK_TYPES, 
    WATCHME_DEFAULT_TYPE
)
import watchme
import argparse
import sys
import os


def get_parser():
    parser = argparse.ArgumentParser(description="WatchMe Command Line Tool")

    # Global Variables

    parser.add_argument('--debug', dest="debug", 
                        help="use verbose logging to debug.", 
                        default=False, action='store_true')

    parser.add_argument('--version', dest="version", 
                        help="show version and exit.", 
                        default=False, action='store_true')

    parser.add_argument('--quiet', dest="quiet", 
                        help="suppress additional output.", 
                        default=False, action='store_true')

    parser.add_argument('--watcher', dest="watcher", 
                        help="the watcher to create (defaults to watcher)", 
                        default=None, type=str)

    parser.add_argument('--base', dest="base", 
                        help="the watcher base (defaults to $HOME/.watchme)", 
                        default=None, type=str)

    description = 'actions for WatchMe Command Line Tool'

    subparsers = parser.add_subparsers(help='watchme actions',
                                       title='actions',
                                       description=description,
                                       dest="command")

    # init

    init = subparsers.add_parser("init",
                                 help="initialize watchme")
 
    init.add_argument('--empty', dest="create_empty", 
                      help="don't create the default watcher folder", 
                      default=False, action='store_true')

    # get

    get = subparsers.add_parser("get",
                                help="get a watcher repository or task using git")

    get.add_argument('repo', nargs=1,
                     help='the repository url to clone')

    get.add_argument('--force', dest="force", 
                     help="force overwrite a watcher, if already exists.", 
                     default=False, action='store_true')

    # export

    export = subparsers.add_parser("export",
                                  help="export data for a task and result file")

    export.add_argument('watcher', nargs=1,
                         help='the watcher export data from')

    export.add_argument('task', nargs=1,
                         help='the name of the task to export data from')

    export.add_argument('filename', nargs=1,
                         help='the filename in the task folder to export')

    export.add_argument('--force', dest="force", 
                         help="force overwrite output, if already exists.", 
                         default=False, action='store_true')

    export.add_argument('--json', dest="json", 
                         help="signal to load the file as json", 
                         default=False, action='store_true')

    export.add_argument('--out', dest="out", 
                         help="the output file (json) to export the data",
                         default=None)


    # create

    create = subparsers.add_parser("create",
                                   help="create a new watcher")

    create.add_argument('watchers', nargs="*",
                        help='watchers to create (default: single watcher)')

    # add

    add = subparsers.add_parser("add",
                                 help="add a task to a watcher.")

    add.add_argument('watcher', nargs=1,
                     help='the watcher to add to')

    add.add_argument('task', nargs=1,
                     help='the name of the task to add. Must start with task')

    add.add_argument('--type', dest="watcher_type",
                     choices=WATCHME_TASK_TYPES, 
                     default=WATCHME_DEFAULT_TYPE)

    add.add_argument('--active', dest="active",
                     choices=["true", "false"], 
                     default="true")

    add.add_argument('--force', dest="force", 
                     help="force overwrite a task, if already exists.", 
                     default=False, action='store_true')


    # inspect

    inspect = subparsers.add_parser("inspect",
                                    help="inspect a task or watcher")

    inspect.add_argument('watcher', nargs=1,
                         help='the watcher to inspect')

    inspect.add_argument('--add-command', dest="create_command", 
                         help="print the command used to add a a task.", 
                         default=False, action='store_true')

    # list

    ls = subparsers.add_parser("list",
                               help="list all watchers at a base")

    # protect and freeze

    protect = subparsers.add_parser("protect",
                                    help="protect or freeze a watcher.")

    protect.add_argument('watcher', nargs=1,
                          help='the watcher to protect or freeze')

    protect.add_argument('action',
                         default="on",
                         choices=['on', 'off', 'freeze', 'unfreeze'])

    # remove

    remove = subparsers.add_parser("remove",
                                   help="remove a task or entire watcher.")

    remove.add_argument('watcher', nargs=1,
                        help='the watcher to remove tasks from')

    remove.add_argument('--delete', dest="delete", 
                        help="delete the entire watch repository", 
                        default=False, action='store_true')

    # run

    run = subparsers.add_parser("run",
                                help="run a watcher.")

    run.add_argument('watcher', nargs=1,
                     help='the watcher to run')

    run.add_argument('--serial', dest="serial", 
                     help="run tasks in serial", 
                     default=False, action='store_true')

    run.add_argument('--test', dest="test", 
                     help="run tasks in test mode (no save)", 
                     default=False, action='store_true')

    run.add_argument('--no-progress', dest="no_progress", 
                     help="run tasks in serial", 
                     default=False, action='store_true')

    # activate

    activate = subparsers.add_parser("activate",
                                     help="activate a new watcher")

    activate.add_argument('watcher', nargs=1,
                          help='watcher to activate (optionally add tasks)')

    # deactivate

    deactivate = subparsers.add_parser("deactivate",
                                       help="deactivate a watcher")

    deactivate.add_argument('watcher', nargs=1,
                            help='watcher to deactivate (optionally add tasks)')


    # schedule

    schedule = subparsers.add_parser("schedule",
                                     help="schedule your watcher.")
 
    schedule.add_argument('watcher', nargs=1,
                           help='the watcher to schedule')

    schedule.add_argument('--force', dest="force", 
                           help="force overwrite of a schedule, if exists.", 
                           default=False, action='store_true')

    # edit

    edit = subparsers.add_parser("edit",
                                  help="edit a watcher task.")
 
    edit.add_argument('watcher', nargs=1,
                      help='the watcher to edit')

    edit.add_argument('action', nargs=1, 
                      help="the action to take",
                      choices=['add', 'update', 'remove'])

    edit.add_argument('task', nargs=1, 
                      help="the task to edit.")

    return parser


def get_subparsers(parser):
    '''get_subparser will get a dictionary of subparsers, to help with printing help
    '''

    actions = [action for action in parser._actions 
               if isinstance(action, argparse._SubParsersAction)]

    subparsers = dict()
    for action in actions:
        # get all subparsers and print help
        for choice, subparser in action.choices.items():
            subparsers[choice] = subparser

    return subparsers



def main():
    '''the main entry point for the WatchMe Command line application.
    '''

    # Customize parser

    parser = get_parser()
    subparsers = get_subparsers(parser)

    def help(return_code=0):
        '''print help, including the software version and active client 
           and exit with return code.
        '''

        version = watchme.__version__
        
        bot.custom(message='Command Line Tool v%s' % version,
                   prefix='\n[WatchMe] ', 
                   color='CYAN')

        parser.print_help()
        sys.exit(return_code)
    
    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()
    try:
        args, unknown = parser.parse_known_args()
    except:
        sys.exit(0)

    extras = None
    if len(unknown) > 0:
        extras = unknown

    # if environment logging variable not set, make silent

    if args.debug is False:
        os.environ['MESSAGELEVEL'] = "INFO"

    # Show the version and exit
    if args.version is True:
        print(watchme.__version__)
        sys.exit(0)

    if args.command == "activate": from .activate import main
    elif args.command == "add": from .add import main
    elif args.command == "edit": from .edit import main
    elif args.command == "export": from .export import main
    elif args.command == "create": from .create import main
    elif args.command == "deactivate": from .deactivate import main
    elif args.command == "get": from .get import main
    elif args.command == "init": from .init import main
    elif args.command == "inspect": from .inspect import main
    elif args.command == "list": from .ls import main
    elif args.command == "protect": from .protect import main
    elif args.command == "remove": from .remove import main
    elif args.command == "run": from .run import main
    elif args.command == "schedule": from .schedule import main
    else: help()

    # Pass on to the correct parser
    return_code = 0
    try:
        main(args, extras)
        sys.exit(return_code)
    except UnboundLocalError:
        return_code = 1

    help(return_code)

if __name__ == '__main__':
    main()



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.defaults import WATCHME_BASE_DIR
from watchme.utils import ( get_tmpdir, run_command )
from watchme.logger import bot

import os
import re
import shutil
import sys

def get_watchers(base=None, quiet=False):
    '''list the watchers installed at a base. If base is not defined,
       the default base is used.

       Parameters
       ==========
       base: the watchme base, defaults to $HOME/.watchme
    '''
    if base == None:
        base = WATCHME_BASE_DIR

    if os.path.exists(base):
        watchers = os.listdir(base)
        if quiet == False:
            bot.info('\n'.join(watchers))
        return watchers
    else:
        bot.exit('%s does not exist.' % base)


def list_watcher(watcher, base=None):
    '''list the contents (tasks) of a single watcher.

       Parameters
       ==========
       base: the watchme base, defaults to $HOME/.watchme
    '''
    if base == None:
        base = WATCHME_BASE_DIR

    repo = os.path.join(base, watcher)
    if os.path.exists(repo):
        files = os.listdir(repo)
        bot.custom(prefix="task:", message="%s" % repo, color="CYAN")
        bot.info('\n  '.join(files))
    else:
        bot.exit('%s does not exist.' % base)

def clone_watcher(repo, base=None, name=None):
    '''clone a watcher from Github (or other version control with git)
       meaning that we clone to a temporary folder, and then move
       to a new folder. By default, the user gets all tasks associated
       with the watcher, along with the git folder so that removing
       is also done with version control.

       Parameters
       ==========
       repo: the repository to clone
       base: the watchme base, defaults to $HOME/.watchme
       name: a new name for the watcher, if a rename is desired.
    '''    
    if base == None:
        base = WATCHME_BASE_DIR

    # clone_watcher(repo=repo, base=args.base, name=extra)
    # STOPPED HERE - need to test this.

    # Validate the repository address
    if not re.search('^git@|http', repo):
        bot.exit('Please provide a valid url to git repository')

    # if the name is None, use the repo name
    if name == None:
        name = os.path.basename(repo) 

    # Ensure we aren't overwriting
    dest = os.path.join(base, name)
    if os.path.exists(dest):
        bot.exit('%s already exists, choose a different watcher name.' % name)

    clone_dest = get_tmpdir(prefix="watchme-clone", create=False)
    run_command("git clone %s %s" %(repo, clone_dest))

    # Valid by default - will copy over if valid
    valid = True    

    # Iterate over watchers
    watchers = os.listdir(clone_dest)
    for watcher in watchers:
        watcher = os.path.join(clone_dest, watcher)
        tasks = os.listdir(watcher)

        # Check that tasks include watchme.cfg
        for task in tasks:
            if not task.startswith('task'):
                continue
            task_folder = os.path.join(watcher, task)
            content = os.listdir(task_folder)
            if 'watcher.cfg' not in content:
                bot.error('%s is missing a watcher.cfg' % task)
                valid = False 
                break
                
    if valid:
        shutil.move(clone_dest, dest)
        
    if os.path.exists(clone_dest):
        shutil.rmtree(clone_dest)



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme.logger import ( bot, RobotNamer )
from watchme.version import __version__
from watchme.defaults import (
    WATCHME_BASE_DIR,
    WATCHME_TASK_TYPES,
    WATCHME_NOTALLOWED_PARAMS
)

from watchme.command import (
    create_watcher,
    write_timestamp,
    get_watchers,
    git_commit,
    git_add
)

from configparser import NoOptionError

from .data import (
    export_dict
)

from .settings import (
    get_setting,
    set_setting,
    get_section,
    print_section,
    print_add_task,
    remove_setting,
    remove_section
)

from .schedule import (
    remove_schedule,
    get_crontab,
    get_job,
    has_schedule,
    update_schedule,
    clear_schedule,
    schedule
)

from watchme.config import (
    read_config,
    write_config
)

from watchme.utils import (
    mkdir_p,
    write_file,
    write_json
)

import os
import re
import shutil
import json
import sys


class Watcher(object):

    repo=None
    configfile=None

    def __init__(self, name=None, 
                       base=None, 
                       create=False, **kwargs):
        '''the watcher base loads configuration files for the user (in $HOME)
           and module, and then stores any arguments given from the caller

           Parameters
           ==========
           name: the watcher name, defaults to github
           base: the watcher base, will default to $HOME/.watchme
           create: boolean to create the watcher if doesn't exist (default False)
           kwargs: should include command line arguments from the client.

        '''
        # Set the watcher base
        self._set_base(base, create)
        self._version = __version__

        # Load the configuration
        self.load_config()


    def _set_base(self, base=None, create=False):
        ''' set the base for the watcher, ensuring that it exists.

            Parameters
            ==========
            base: the base folder of watcher repos. Uses $HOME/.watchme default
            create: create the watcher if it doesn't exist (default is False)
        '''
        if base == None:
            base = WATCHME_BASE_DIR

        # Does the watcher exist?
        self.base = base
        self.repo = os.path.join(self.base, self.name)
        self.configfile = os.path.join(self.repo, 'watchme.cfg')

        # If the watcher doesn't exist and we need to create:
        if not os.path.exists(self.repo) or not os.path.exists(self.configfile):
            if create is True:
                create_watcher(self.name)   
            else:
                bot.exit('Watcher %s does not exist. Use watchme create.' % self.name)


# Config

    def save(self):
        '''save the configuration to file.'''
        write_config(self.configfile, self.config)


    def edit_task(self, name, action, key, value=None):
        '''edit a task, meaning doing an addition (add), update (update), or
           "remove", All actions require a value other than remove.

           Parameters
           ==========
           name: the name of the task to update
           action: the action to take (update, add, remove) a parameter
           key: the key to update
           value: the value to update
        '''

        if not self.has_task(name):
            bot.exit('%s is not a task defined by %s' %(name, self.name))

        if action not in ['update', 'add', 'remove']:
            bot.exit('Action must be update, add, or remove')

        if action in ['update', 'add'] and value == None:
            bot.exit('A value must be provided for the %s action' % action)

        # Add, and it doesn't exist so it's okay
        if action == "add" and key not in self.config[name]:
            bot.info('Adding %s:%s to %s' %(key, value, name))
            self.set_setting(name, key, value)

        # Already exists, encourage user to update
        elif action == "add" and key in self.config[name]:
            bot.exit('%s already exists. Use "update" action to change.' % key)

        # Update, and it's a valid choice
        elif action == 'update' and key in self.config[name]:
            bot.info('Updating %s to %s in %s' %(key, value, name))
            self.set_setting(name, key, value)

        # Update, and it's not a valid choice
        elif action == 'update' and key not in self.config[name]:
            bot.exit('%s is not found in config, cannot be updated.' % key)

        # Remove, and it's a valid choice
        elif action == "remove" and key in self.config[name]:
            bot.info('Removing %s' % key )
            del self.config[name][key]

        # Remove, and it's not a valid choice
        elif action == "remove" and key not in self.config[name]:
            bot.exit('%s is not found in config, cannot be removed.' % key)
        self.save()


    def has_section(self, name):
        '''returns True or False to indicate if the watcher has a specified
           section. To get a task, use self.has_task.

           Parameters
           ==========
           name: the name of the section to check for.
        '''
        self.load_config()
        if name in self.config._sections:
            return True
        bot.warning('%s not found for watcher %s' %(name, self.name))
        return False


    def load_config(self):
        '''load a configuration file, and set the active setting for the watcher
           if the file doesn't exist, the function will exit and prompt the user 
           to create the watcher first. If the watcher section isn't yet defined,
           it will be written with a default active status set to false.
        '''
        if not hasattr(self, 'config'):
           
            # Load the configuration file if it exists (will exit if not found)
            if self.configfile != None:
                self.config = read_config(self.configfile)

            # The watcher section is added by default, args for the watcher
            if 'watcher' not in self.config.sections():
                self.config.add_section('watcher')
                self.set_setting('watcher', 'active', 'false')

                # Only update the config if we've changed it
                self.save()


    def _get_params_dict(self, pairs):
        '''iterate through parameters, make keys lowercase, and ensure
           valid format.

           Parameters
           ==========
           pairs: a list of key@value pairs to set.
        '''
        params = {}
        for pair in pairs:
            if "@" not in pair:
                bot.exit('incorrectly formatted param, must be key@value')
            key,value = pair.split('@', 1)
            key = key.lower()

            # All tasks are not allowed to have default params
            if key in WATCHME_NOTALLOWED_PARAMS:
                bot.error('%s is a default, not allowed setting by task.' % key)
                self.valid = False
            params[key] = value
        return params


# Add Tasks

    def add_task(self, task, task_type, params, force=False, active="true"):
        '''add a task, meaning ensuring that the type is valid, and that
           the parameters are valid for the task.

           Parameters
           ==========
           task: the Task object to add, should have a name and params and
                 be child of watchme.tasks.TaskBase
           task_type: must be in WATCHME_TASK_TYPES, meaning a client exists
           params: list of parameters to be validated (key@value)
           force: if task already exists, overwrite
           active: add the task as active (default "true")
           
        '''
        # Check again, in case user calling from client
        if not task.startswith('task'):
            bot.exit('Task name must start with "task" (e.g., task-reddit)')
       
        # Ensure it's a valid type
        if task_type not in WATCHME_TASK_TYPES:
            bot.exit('%s is not a valid type: %s' % WATCHME_TASK_TYPES)

        # Validate variables provided for task
        if task_type.startswith('url'):
            from .urls import Task

        # Validate variables provided for task
        elif task_type == 'psutils':
            from .psutils import Task

        else:
            bot.exit('task_type %s not properly added to Watcher' % task_type)

        # Convert list to dictionary
        params = self._get_params_dict(params)
 
        # Creating the task will validate parameters
        newtask = Task(task, params=params)

        # Exit if the new task is not valid
        if not newtask.valid:
            bot.exit('%s is not valid, will not be added.' % task)

        # Write to file (all tasks get active = True added, and type)
        self._add_task(newtask, force, active)


    def _add_task(self, task, force=False, active='true'):
        '''add a new task to the watcher, meaning we:

           1. Check first that the task doesn't already exist (if the task
              exists, we only add if force is set to true)
           2. Validate the task (depends on the task)
           3. write the task to the helper config file, if valid.

           Parameters
           ==========
           task: the Task object to add, should have a name and params and
                 be child of watchme.tasks.TaskBase
           force: if task already exists, overwrite
           active: add the task as active (default "true")
        '''
        self.load_config()

        if active not in ["true", "false"]:
            bot.exit('Active must be "true" or "false"')

        # Don't overwrite a section that already exists
        if task.name in self.config.sections():
            if not force:
                bot.exit('%s exists, use --force to overwrite.' % task.name)
            self.remove_section(task.name, save=False)

        # Add the new section
        self.config[task.name] = task.export_params(active=active)
        self.print_section(task.name)
        self.save()

        # If the task folder doesn't exist, recreate it.
        task_folder = os.path.join(self.repo, task.name)
        if not os.path.exists(task_folder):
            mkdir_p(task_folder)
            git_add(self.repo, task.name)

        # Commit changes
        git_commit(repo=self.repo, task=self.name, message="ADD task %s" % task.name)

# Delete

    def delete(self):
        '''delete the entire watcher, only if not protected. Cannot be undone.
        '''
        self.load_config()

        # Check for protection
        if self.is_frozen():
            bot.exit('watcher %s is frozen, unfreeze to delete.' % self.name)
        elif self.is_protected():
            bot.exit('watcher %s is protected, turn off protection to delete.' % self.name)

        repo = os.path.dirname(self.configfile)

        # Ensure repository exists before delete
        if os.path.exists(repo):
            bot.info('Removing watcher %s' % self.name)
            shutil.rmtree(repo)
        else:
            bot.exit("%s:%s doesn't exist" %(self.name, repo))


    def remove_task(self, task):
        '''remove a task from the watcher repo, if it exists, and the
           watcher is not frozen.

           Parameters
           ==========
           task: the name of the task to remove
        '''
        if self.get_section(task) != None:
            if self.is_frozen():
                bot.exit('watcher is frozen, unfreeze first.')
            self.remove_section(task)

            # If the task has a folder, remove the entire thing
            repo = os.path.join(self.repo, task)
            if os.path.exists(repo):
                shutil.rmtree(repo)

            bot.info('%s removed successfully.' % task)
            git_commit(self.repo, self.name, "REMOVE task %s" % task)

        else:
            bot.warning('Task %s does not exist.' % task)


# Inspect
    
    def inspect(self, tasks=None, create_command=False):
        '''inspect a watcher, or one or more tasks belonging to it. This means
           printing the configuration for the entire watcher (if tasks is None)
           or just for one or more tasks.
 
           Parameters
           ==========
           tasks: one or more tasks to inspect (None will show entire file)
           create_command: if True, given one or more tasks, print the command
                           to create them.
        '''
        self.load_config()
        if tasks == None:
            tasks = self.config.sections()

        # If the user supplied one task:
        if not isinstance(tasks, list):
            tasks = [tasks]

        # Show all sections
        for task in tasks:

            # If the user doesn't want to see the create command:
            if create_command is False:
                self.print_section(task)
                bot.newline()
            else:
                self.print_add_task(task)


    def list(self, quiet=False):
        '''list the watchers. If quiet is True, don't print to the screen.'''
        watchers = get_watchers(base=self.base, quiet=quiet)
        return watchers

# Protection

    def protect(self, status="on"):
        '''protect a watcher, meaning that it cannot be deleted. This does
           not influence removing a task. To freeze the entire watcher,
           use the freeze() function.
        '''        
        self._set_status('watcher', 'protected', status)
        git_commit(self.repo, self.name, "PROTECT %s" % status)
        self.print_section('watcher')


    def freeze(self):
        '''freeze a watcher, meaning that it along with its tasks cannot be 
           deleted. This does not prevent the user from manual editing.
        '''
        self._set_status('watcher', 'frozen', 'on')
        git_commit(self.repo, self.name, "FREEZE")
        self.print_section('watcher')

    def unfreeze(self):
        '''freeze a watcher, meaning that it along with its tasks cannot be 
           deleted. This does not prevent the user from manual editing.
        '''
        self._set_status('watcher', 'frozen', 'off')
        git_commit(self.repo, self.name, "UNFREEZE")
        self.print_section('watcher')

    def _set_status(self, section, setting, value):
        '''a helper function to set a status, ensuring that status value
           is in "on" or "off"

           Parameters
           ==========
           status: one of "on" or "off"
           name: a value to set status for.
        '''
        if value not in ['on', 'off']:
            bot.exit('Status must be "on" or "off"')
        self.set_setting(section, setting, value)
        self.save()
        

    def is_protected(self):
        '''return a boolean to indicate if the watcher is protected or frozen.
           protected indicates no delete to the watcher, but allowed delete
           to tasks, frozen indicates no change of anything.
        '''
        protected = False
        for status in ['protected', 'frozen']:
            if self.get_setting('watcher', status) == "on":
                protected = True
        return protected
        

    def is_frozen(self):
        '''return a boolean to indicate if the watcher is frozen.
           protected indicates no delete to the watcher, but allowed delete
           to tasks, frozen indicates no change of anything.
        '''
        if self.get_setting('watcher', 'frozen') == "on":
            return True
        return False

# Status

    def _active_status(self, status='true', name=None):
        '''a general function to change the status, used by activate and
           deactivate.
 
           Parameters
           ==========
           status: must be one of true, false
           name: if not None, we are deactivating a task (not the watcher)
        '''
        # Load the configuration, if not loaded
        self.load_config()

        if name == None:
            name = 'watcher'

        # Cut out early if section not in config
        if name not in self.config._sections:
            bot.exit('%s is not a valid task or section' % name)     

        if status not in ['true', 'false']:
            bot.exit('status must be true or false.')

        # Update the status and alert the user
        self.set_setting(name, 'active', status)
        self.save()

        # Return the message for the commit
        message = "ACTIVE"
        if status == "false":
            message = "DEACTIVATE"

        # Add the task name
        if name != None:
            message = "%s task %s" %(message, name)

        bot.info('[%s|%s] active: %s' % (name, self.name, status)) 
        return message

    def activate(self, task=None):
        '''turn the active status of a watcher to True
        '''
        message = self._active_status('true', task)
        git_commit(self.repo, self.name, message)
   

    def deactivate(self, task=None):
        '''turn the active status of a watcher to false. If a task is provided,
           update the config value for the task to be false.
        '''
        # If no task defined, user wants to deactiate watcher
        message = self._active_status('false', task)
        git_commit(self.repo, self.name, message)


    def is_active(self, task=None):
        '''determine if the watcher is active by reading from the config directly
           if a task name is provided, check the active status of the task
        '''
        if task == None:
            task = 'watcher'
        if self.get_setting(task, 'active', default='true') == "true":
            return True
        return False


# Get and Prepare Tasks

    def has_task(self, name):
        '''returns True or False to indicate if the watcher has a specified
           task.
        '''
        self.load_config()
        if self.has_section(name) and name.startswith('task'):
            return True
        return False


    def get_task(self, name, save=False):
        '''get a particular task, based on the name. This is where each type
           of class should check the "type" parameter from the config, and
           import the correct Task class.

           Parameters
           ==========
           name: the name of the task to load
           save: if saving, will be True
        '''
        self.load_config()

        task = None

        # Only sections that start with task- are considered tasks
        if name in self.config._sections and name.startswith('task'):

            # Task is an ordered dict, key value pairs are entries
            params = self.config._sections[name]

            # Get the task type (if removed, consider disabled)
            task_type = params.get('type', '')

            # If we get here, validate and prepare the task
            if task_type.startswith("url"):
                from .urls import Task

            elif task_type == 'psutils':
                from .psutils import Task

            else:
                bot.exit('Type %s not properly set up in get_task' % task_type)

            # if not valid, will return None
            task = Task(name, params, _save=save)

        return task


    def _task_selected(self, task, regexp=None):
        '''check if a task is active and (if defined) passes user provided
           task names or regular expressions.

           Parameters
           ==========
           task: the task object to check
           regexp: an optional regular expression (or name) to check
        '''
        selected = True 

        # A task can be None if it wasn't found
        if task == None:
            selected = False

        # Is the task not active (undefined is active)?
        active = task.params.get('active', 'true')
        if active == "false":
            bot.info('Task %s is not active.' % task)
            selected = False
        
        # The user wants to search for a custom task name
        if regexp != None:
            if not re.search(regexp, task):
                bot.info('Task %s is selected to run.' % task)
                selected = False

        return selected


    def get_tasks(self, regexp=None):
        '''get the tasks for a watcher, possibly matching a regular expression.
           A list of dictionaries is returned, each holding the parameters for
           a task. "uri" will hold the task (folder) name, active

           Parameters
           ==========
           regexp: if supplied, the user wants to run only tasks that match
                   a particular pattern
        '''
        self.load_config()

        tasks = []
        for section in self.config._sections:

            # Get the task based on the section name
            task = self.get_task(section)

            # Check that the task should be run, and is valid
            if task != None:
                if self._task_selected(task, regexp) and task.valid:
                    tasks.append(task)

        bot.info('Found %s contender tasks.' % len(tasks))
        return tasks   


# Running Tasks

    def run_tasks(self, queue, parallel=True, show_progress=True):
        '''this run_tasks function takes a list of Task objects, each
           potentially a different kind of task, and extracts the parameters
           with task.export_params(), and the running function with 
           task.export_func(), and hands these over to the multiprocessing
           worker. It's up to the Task to return some correct function
           from it's set of task functions that correspond with the variables.

           Examples
           ========

           funcs
           {'task-reddit-hpc': <function watchme.watchers.urls.tasks.get_task>}

           tasks
           {'task-reddit-hpc': [('url', 'https://www.reddit.com/r/hpc'),
                                ('active', 'true'),
                                ('type', 'urls')]}
        ''' 
        if parallel is True:
            return self._run_parallel(queue, show_progress)
        
        # Otherwise, run in serial
        results = {}

        # Progressbar
        total = len(queue)
        progress = 1

        for task in queue:
            prefix = "[%s:%s/%s]" % (task.name, progress, total)
            if show_progress is True:
                bot.show_progress(progress, total, length=35, prefix=prefix)
            else:
                bot.info('Running %s' % prefix)
            results[task.name] = task.run()
            progress+=1

        return results


    def _run_parallel(self, queue, show_progress=True):
        ''' run tasks in parallel using the Workers class. Returns a dictionary
            (lookup) wit results, with the key being the task name

            Parameters
            ==========
            queue: the list of task objects to run
        '''
        from watchme.tasks.worker import Workers

        # Run with multiprocessing
        funcs = {}
        tasks = {}

        for task in queue:

            # Export parameters and functions            
            funcs[task.name] = task.export_func()
            tasks[task.name] = task.export_params()

        workers = Workers(show_progress=show_progress)
        return workers.run(funcs, tasks)


    def run(self, regexp=None, parallel=True, test=False, show_progress=True):
        '''run the watcher, which should be done via the crontab, including:

             - checks: the instantiation of the client already ensures that 
                       the watcher folder exists, and has a configuration,
                       and it loads.
             - parse: parse the tasks to be run
             - start: run the tasks that are defined for the watcher.
             - finish: after completion, commit to the repository changed files

           Parameters
           ==========
           regexp: if supplied, the user wants to run only tasks that match
                   a particular pattern         
           parallel: if True, use multiprocessing to run tasks (True)
                     each watcher should have this setup ready to go. 
           test: run in test mode (no saving of results)
           show_progress: if True, show progress bar instead of task information
                          (defaults to True)
        '''
        # Step 0: Each run session is given a fun name
        run_id = RobotNamer().generate()

        # Step 1: determine if the watcher is active.
        if self.is_active() == False and test is False:
            bot.exit('Watcher %s is not active.' % self.name)

        # Step 2: get the tasks associated with the run, a list of param dicts
        tasks = self.get_tasks()

        # Step 3: Run the tasks. This means preparing a list of funcs/params,
        # and then submitting with multiprocessing
        results = self.run_tasks(tasks, parallel, show_progress)

        # Finally, finish the runs.
        if test is False:
            self.finish_runs(results)
        else:
            # or print results to the screen
            print(json.dumps(results, indent=4))


    def finish_runs(self, results):
        '''finish runs should take a dictionary of results, with keys as the
           folder name, and for each, depending on the result type,
           write the result to file (or update file) and then commit
           to git.

           Parameters
           ==========
           results: a dictionary of tasks, with keys as the task name, and
                    values as the result.
        '''
        for name, result in results.items():
            task_folder = os.path.join(self.repo, name)
            task = self.get_task(name, save=True)

            # Files to be added via Git after
            files = []

            # Ensure that the task folder exists
            if not os.path.exists(task_folder):
                mkdir_p(task_folder)
                git_add(self.repo, task_folder)

            # Case 1. The result is a list
            if isinstance(result, list):           

                # Get rid of Nones, if the user accidentally added
                result = [r for r in result if r]

                if len(result) == 0:
                    bot.error('%s returned empty list of results.' % name)

                # json output is specified
                elif task.params.get('save_as') == 'json':
                    bot.debug('Saving single list as one json...')
                    files.append(task._save_json(result, self.repo))

                elif task.params.get('save_as') == 'json':
                    bot.debug('Saving single list as multiple json...')
                    files += task._save_json_list(result, self.repo)

                # Otherwise, sniff for list of paths
                elif os.path.exists(result[0]):
                    bot.debug('Found list of paths...')
                    files += task._save_files_list(result, self.repo)     

                # Finally, assume just writing text to file
                else:
                    bot.debug('Saving content from list to file...')
                    files += task._save_text_list(result, self.repo)     

            # Case 2. The result is a string
            elif isinstance(result, str):

                # if it's a path to a file, just save to repository
                if os.path.exists(result):
                    files.append(task._save_file(result, self.repo))

                # Otherwise, it's a string that needs to be saved to file
                else:
                    files.append(task._save_text(result, self.repo))

            # Case 3. The result is a dictionary
            elif isinstance(result, dict):
                files.append(task._save_json(result, self.repo))

            elif result == None:
                bot.error('Result for task %s is None' % name)

            else:
                bot.error('Unsupported result format %s' % type(result))

            # Get rid of None results (don't check excessively for None above)
            files = [f for f in files if f]

            # Add files to git, and commit
            files.append(write_timestamp(repo=self.repo, task=name))
            git_add(repo=self.repo, files=files)
            git_commit(repo=self.repo, 
                       task=self.name, 
                       message="ADD results %s" % name)


# Identification

    def __repr__(self):
        return "[watcher|%s]" %self.name

    def __str__(self):
        return "[watcher|%s]" %self.name


# Settings

Watcher.remove_setting = remove_setting
Watcher.get_setting = get_setting
Watcher.get_section = get_section
Watcher.set_setting = set_setting
Watcher.remove_section = remove_section
Watcher.print_section = print_section
Watcher.print_add_task = print_add_task

# Schedule 

Watcher.remove_schedule = remove_schedule
Watcher.get_crontab = get_crontab
Watcher.update_schedule = update_schedule
Watcher.has_schedule = has_schedule
Watcher.get_job = get_job
Watcher.clear_schedule = clear_schedule
Watcher.schedule = schedule

# Data

Watcher.export_dict = export_dict



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

import os
import tempfile
import requests

# Helper functions

def get_params(kwargs, key='url_param_'):
    '''a general function to get parameter sets based on a user input. 
       Returns a list of dictionaries, one per set.

       Parameters
       ==========
       kwargs: the dictionary of keyword arguments that may contain url
               parameters (format is url_param_<name>
       key: the string that the parameters start with (defaults to url_param)
    '''
    # Keey dictionary based on index of param
    params = {}

    names = [x for x in kwargs if x.startswith(key)]
    for n in range(len(names)):
        name = names[n]
        # Params are split by commas, with index corresponding to list index
        paramlist = kwargs.get(name).split(',')
        # Remove the "url_param"
        name = name.replace(key, '', 1)
        # Update the dictionary of dictionaries
        for i in range(len(paramlist)):

            # Add the index to the dicionary (will be turned in list later)
            if i not in params:
                params[i] = {}

            if paramlist[i] != '':
                params[i][name] = paramlist[i]

    # Unwrap the list
    params = [x for x in params.values()]

    # If no special params, we at least need to run once
    if len(params) == 0:
        params = [{}]

    return params


def get_headers(kwargs):
    '''Get a single set of headers from the kwargs dict.

       Parameters
       ==========
       kwargs: the dictionary of keyword arguments that may contain url
               parameters (format is url_param_<name>
    '''
    headers = {}

    for key, value in kwargs.items():
        if key.startswith('header_'):
            name = key.replace('header_', '', 1)
            headers[name] = value

    return headers


def get_results(url, selector, func=None, attributes=None, params={}, get_text=False, headers={}):
    '''given a url, a function, an optional selector, optional attributes, and a set (dict)
       of parameters, perform a request.

       Parameters
       ==========
       url: the url to get (required)
       func: the function to use, defaults to requests.get
       selector:  selection for the html response
       attributes: optional, a list of attributes
       params: a dictionary of parameters
       headers: a dictionary of header key value pairs
    '''
    from bs4 import BeautifulSoup

    if func == None:
        func = requests.get

    response = func(url, params=params, headers=headers)
    results = []

    if response.status_code == 200:   
        soup = BeautifulSoup(response.text, 'lxml')

        # Get the selection
        for entry in soup.select(selector):

            # Does the user want to get attributes
            if attributes != None:
                [results.append(entry.get(x)) for x in attributes]

            # Does the user want to get text?
            elif get_text == True:
                results.append(entry.text)

            # Otherwise, return the entire thing
            else:
                results.append(str(entry))

    # Clean up results
    results = [x for x in results if x]

    return results



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

The watcher is actually a connection to crontab. This is what helps to schedule
the watched to check for changes at some frequency, and update the files.


'''

import configparser
import errno
import os
import pwd
import re
import tempfile
import json
import io
import socket
import shutil
import sys

from watchme.logger import bot

# FOLDER OPERATIONS ############################################################

def get_userhome():
    '''get the user home based on the effective uid
    '''
    return pwd.getpwuid(os.getuid())[5]

def get_user():
    '''return the active user'''
    return os.path.basename(get_userhome())

def get_host():
    '''return the hostname'''
    return socket.gethostname()

def mkdir_p(path):
    ''' mkdir_p attempts to get the same functionality as mkdir -p

        Paramters
        =========
        param path: the path to create.
    '''
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            bot.error("Error creating path %s, exiting." % path)
            sys.exit(1)


# FILE OPERATIONS ##############################################################

def generate_temporary_file(folder=None, prefix='watchme', ext=None):
    '''write a temporary file, in base directory with a particular extension.
      
       Parameters
       ==========
       folder: the base directory to write in. 
       prefix: the prefix to use
       ext: the extension to use.

    '''        
    if folder == None:
        folder = tempfile.gettempdir()
    tmp = next(tempfile._get_candidate_names())
    tmp = '%s/%s.%s' %(folder, prefix, tmp)

    # Does the user want an extension?
    if ext != None:
        tmp = "%s.%s" %(tmp, ext)

    return tmp


def get_tmpdir(prefix="", create=True):
    '''get a temporary directory for an operation. If SREGISTRY_TMPDIR
       is set, return that. Otherwise, return the output of tempfile.mkdtemp

       Parameters
       ==========
       prefix: Given a need for a sandbox (or similar), we will need to 
       create a subfolder *within* the SREGISTRY_TMPDIR.
       create: boolean to determine if we should create folder (True)
    '''
    tmpdir = tempfile.gettempdir()
    prefix = prefix or "watchme-tmp"
    prefix = "%s.%s" %(prefix, next(tempfile._get_candidate_names()))
    tmpdir = os.path.join(tmpdir, prefix)

    if not os.path.exists(tmpdir) and create is True:
        os.mkdir(tmpdir)

    return tmpdir


def copyfile(source, destination, force=True):
    '''copy a file from a source to its destination.
    '''
    if os.path.exists(destination) and force is True:
        os.remove(destination)
    shutil.copyfile(source, destination)
    return destination


def write_file(filename, content, mode="w"):
    '''write_file will open a file, "filename" and write content, "content"
    and properly close the file
    '''
    with open(filename, mode) as filey:
        filey.writelines(content)
    return filename


def write_json(json_obj, filename, mode="w", print_pretty=True):
    ''' write_json will (optionally,pretty print) a json object to file
    
        Parameters
        ==========
        json_obj: the dict to print to json
        filename: the output file to write to
        pretty_print: if True, will use nicer formatting
    '''
    with open(filename, mode) as filey:
        if print_pretty:
            filey.writelines(print_json(json_obj))
        else:
            filey.writelines(json.dumps(json_obj))
    return filename


def print_json(json_obj):
    ''' just dump the json in a "pretty print" format
    '''
    return json.dumps(
                    json_obj,
                    indent=4,
                    separators=(
                        ',',
                        ': '))


def read_file(filename, mode="r", readlines=True):
    '''write_file will open a file, "filename" and write content, "content"
       and properly close the file
    '''
    with open(filename, mode) as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def read_json(filename, mode='r'):
    '''read_json reads in a json file and returns
       the data structure as dict.
    '''
    with open(filename, mode) as filey:
        data = json.load(filey)
    return data



'''

Copyright (C) 2019 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

'''

from watchme import get_watcher
from watchme.utils import ( write_json, generate_temporary_file )
from watchme.logger import bot
import json
import os

def main(args, extra):
    '''activate one or more watchers
    '''    
    # Required - will print help if not provided
    name = args.watcher[0]
    task = args.task[0]
    filename = args.filename[0]

    if not task.startswith('task'):
        example = 'watchme add watcher task-reddit url@https://www.reddit.com'
        bot.exit('Task name must start with "task", e.g., %s' % example)

    # Use the output file, or a temporary file
    out = args.out

    # Get the watcher to interact with, must already exist
    watcher = get_watcher(name, base=args.base, create=False) 

    if out is not None:
        if os.path.exists(out) and args.force is False:
            bot.exit('%s exists! Use --force to overwrite.' % out)

    # Export the data to file
    result = watcher.export_dict(task=task,
                                 filename=filename,
                                 name=name,
                                 export_json=args.json,
                                 base=args.base)

    if result != None:

        if out == None:
            print(json.dumps(result, indent=4))
        else:
            write_json(result, out)
            bot.info('Result written to %s' % out)
